/*
	copyright 2006 Gilles GIGAN (gilles.gigan@gmail.com)
			
	This file is part of light_cap.

   	light_cap is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    light_cap is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#include <pthread.h>
#include <signal.h>
#include <assert.h>
#include <string.h>			//strstr
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <jpeglib.h>
#include <time.h>			//for nanosleep
#include <sys/types.h> 
#include <sys/socket.h>
#include <sys/time.h> 		//gettimeofday
#include <netinet/in.h>
#include <arpa/inet.h>		//for inet_ntoa (convert strcut in_addr to char*)

#include "v4l2-input.h"
#include "utils.h"
#include "log.h"


#ifdef DEBUG
int log_source=DEFAULT_LOG_SOURCE, log_level = DEFAULT_LOG_LEVEL;
#endif

/* thread-safe ints:
 * client_nr: number of clients connected to the webcam stream
 * jpeg_quality: the requested jpeg quality (0-100%)
 * requested_fps: the requested frame rate (1-25)
 * max_fps: the maximum fps achieved during the calibration period, defined as
 * 			the first SHOW_FPS_INTERVAL seconds after the first client connects
 * 			where frames are sent as fast as possible. max_fps is then measured
 * keep_going: flag used to know when to terminate the capture
 * */
volatile sig_atomic_t client_nr = 0, jpeg_quality, requested_fps, max_fps = 0, keep_going = 1;

//What to print on terminal
int verbosity = LOG_ALL;

/* fps adjustment variables: 
 * fps_nanosleep and fps_secsleep: specify the amount of time(in nanosec and sec resp.) to sleep before 
 * 			capturing a frame. This allows control over the frame rate, between 
 * 			max_fps (fps_secsleep = fps_nanosecsleep = 0) and 1 (frame/sec) 
 * fps_nanosleepstep: specifies an amount of nanosecs used to adjust fps_nanosleep in order to achieve
 * 			requested_fps. fps_nanosleepstep is calculated during the calibration period and is set
 * 			to the time (in ns) required to capture, jpeg-encode and send 1 frame.
 * */
long fps_nanosleep = 0, fps_secsleep = 0, fps_nanosleep_step = 0;;


/* set keep_going to 0 when SIGINT is received
 */
void catch_int (int sig) {
       keep_going = 0;
       signal (sig, catch_int);
}

int main(int argc, char **argv) {
	/*Assumes args are in order !!!!
	1: V4L2 dev name, 2: height, 3: width, 4: channel
	5: standard, 6: jpeg_quality, 7: verbosity, 8: log source, 9: log_level
	*/
	struct capture_device *cdev;
	int sockfd;
	jpeg_quality = JPEG_QUALITY;
	
	//catch ctrl-c
	signal (SIGINT, catch_int);
	signal (SIGPIPE, SIG_IGN );

	//Parse cmd line options
	if(argc != NB_ARGS) {
		print_usage();
		exit(1);
	}

	//init libv4l2 and create cdev struct
	cdev = init_libv4l2(argv[1], atoi(argv[2]), atoi(argv[3]), atoi(argv[4]), atoi(argv[5]), 5);
	jpeg_quality = atoi(argv[6]);
	requested_fps = atoi(argv[7]);
	verbosity = atoi(argv[8]);

#ifdef DEBUG
	log_source = atoi(argv[9]);
	log_level = atoi(argv[10]);
#endif
	
	
	//Open device file
	if (open_device(cdev) < 0) {
		info(LOG_ERR, "Failed to open video device\n");
		exit(1);
	}

	//Check capture & image format capabilities
	if(check_capture_capabilities(cdev) != 0) {
		info(LOG_ERR, "No V4L2 capture capabilities\n");
		exit(1);
	}
	
	//Set capture param (image format, color, size, crop...)
	if (set_cap_param(cdev)!=0) {
		info(LOG_ERR, "Unable to set capture parameters\n");
		exit(1);
	}
	
	info(LOG_INFO, "Capturing at %dx%d\n", cdev->width, cdev->height);

	//Prepare capture:Allocates v4l2 buffers, mmap buffers, enqueue buffers
	if (init_capture(cdev)) {
		info(LOG_ERR, "Failed to setup capture\n");
		exit(1);
	}
	
	//init tcp server
	sockfd = setup_tcp_server_sock(TCP_PORT);
		
	//main loop
	main_loop(sockfd, cdev);

	//close tcp socket
	close(sockfd);

	//Deallocates V4l2 buffers
	free_capture(cdev);
	
	//close device file
	close_device(cdev);

	//delete cdev
	del_libv4l2(cdev);
	
	return 0;
}
/* Sets up a listening (server) socket on port "port"
 * returns the socket descriptor
 */ 
int setup_tcp_server_sock(int port) {
	int sockfd, oldflags;
	struct sockaddr_in serv_addr;
	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	
	if (sockfd < 0) {
    	info(LOG_ERR, "Cant open a socket\n");
		keep_going = 0;
	}
    
    oldflags = 1;
    setsockopt( sockfd, SOL_SOCKET, SO_REUSEADDR, &oldflags, sizeof( oldflags ) ); 
    
    //oldflags = fcntl (sockfd, F_GETFL, 0);
	//fcntl(sockfd, F_SETFL, oldflags | O_NONBLOCK);
        	     
	memset(&serv_addr, 0, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = INADDR_ANY;
	serv_addr.sin_port = htons(port);
	if (bind(sockfd, (struct sockaddr *) &serv_addr,sizeof(serv_addr)) < 0) { 
		info(LOG_ERR,  "Cant bind to if/port\n");
		keep_going = 0;
	}
	listen(sockfd,5);
	return sockfd;
}

/* main loop: waits for incoming TCP connections on "serv_sockfd", and service them
 * cdev is the initialised V4L2 device
 */
void main_loop(int serv_sockfd, struct capture_device *cdev) {
	int clilen, newsockfd;
	struct sockaddr_in cli_addr;
	fd_set rfds;
	
	//Start capture
	if (start_capture(cdev) !=0 )
		keep_going = 0;
	
	clilen = sizeof(cli_addr);
	while(keep_going) {
		FD_ZERO(&rfds);
		FD_SET(serv_sockfd, &rfds);
		
		//select is used so we can sleep and be interrupted while we wait (accept) for an incoming cx
		//using non blocking socket was not good because the while loop would use 100% CPU while accept'ing
		if (select(serv_sockfd+1, &rfds, NULL, NULL, NULL)>0) {
			newsockfd = accept(serv_sockfd, (struct sockaddr *) &cli_addr, (unsigned int *) &clilen);
			info(LOG_INFO, "New connection from %s:%d on socket %d\n", inet_ntoa(cli_addr.sin_addr), cli_addr.sin_port, newsockfd );
			if(newsockfd > 0) {
				switch(get_action(newsockfd, cdev)){
					case ACTION_CAPTURE:
						if(client_nr < MAX_CLIENTS) {
			    	 		//send stream
			     			start_thread_client(newsockfd, cdev);
						}
			 			else {
							info(LOG_ERR, "Cant accept anymore clients.try increasing MAX_CLIENT in main.c and recompile\n");
							close(newsockfd);
						}
						break;
					case ACTION_LIST:
						//output current capture params
						list_cap_param(newsockfd, cdev);
						info(LOG_INFO, "closing connection on socket %d\n", newsockfd );
						close(newsockfd);
						break;
					default:
						break;
				} //end switch
			} //end if new connection
		} //end if(select>0)
	} //while(keep_going)

	//Stop capture
	stop_capture(cdev);
}

/* send an HTMl page over the "newsockfd" socket with a list of V4L2 controls (brightness, contrast, ...)
 * as well as JPEG and fps controls
 */ 
void list_cap_param(int newsockfd,struct capture_device *cdev) {
	char *page, *ptr;
	struct control_list *l;
	int i;

	XMALLOC(page, char *, PARAM_PAGE_SIZE);
	ptr = page;
	ptr += sprintf(ptr,"<html><body>\n");

	//outputs JPEG quality control setting
	ptr += sprintf(ptr,"<form method=\"get\"><h4>JPEG quality</h4>Value: %d (min: 0, max: 100, step: 1)<br>\n", jpeg_quality);
	ptr += sprintf(ptr,"<input type=\"text\" name=\"val\" value=\"%d\" size=\"5\"> &nbsp; <input type=\"submit\" name=\"-1\" value=\"update\"></form>", jpeg_quality);
	write(newsockfd, page, strlen(page));
	ptr = page;
	memset(page, 0, PARAM_PAGE_SIZE);

	//outputs frame rate control setting
	ptr += sprintf(ptr,"<form method=\"get\"><h4>Frame rate</h4>Value: %d (min: 1, max: 25, step: 1)<br>\n", requested_fps);
	ptr += sprintf(ptr,"<input type=\"text\" name=\"val\" value=\"%d\" size=\"3\"> &nbsp; <input type=\"submit\" name=\"-2\" value=\"update\"></form>", requested_fps);
	write(newsockfd, page, strlen(page));
	ptr = page;
	memset(page, 0, PARAM_PAGE_SIZE);

	//outputs controls
	l = cdev->ctrls;
	for(i = 0; i< l->count; i++) {
		ptr += sprintf(ptr,"<form method=\"get\"><h4>%s</h4>Value: %d (min: %d, max: %d, step: %d)<br>\n", l->ctrl[i].name, get_control_value(cdev, &l->ctrl[i]) , l->ctrl[i].minimum, l->ctrl[i].maximum, l->ctrl[i].step);
		ptr += sprintf(ptr,"<input type=\"text\" name=\"val\" value=\"%d\" size=\"5\"> &nbsp; <input type=\"submit\" name=\"%d\" value=\"update\"></form>", get_control_value(cdev, &l->ctrl[i]), i);
		write(newsockfd, page, strlen(page));
		ptr = page;
		memset(page, 0, PARAM_PAGE_SIZE);		
	}
	
	//output end of HTML body
	ptr += sprintf(ptr,"</body></html>\n");
	write(newsockfd, page, strlen(page));
	XFREE(page);
}
/* Reads the first few bytes of "sock" socket and decides what to do (send webcam stream or list of controls)
 * In the latter case (list of controls), more bytes are parsed to see whether we should also set a new value to one
 * of the controls
 */
int get_action(int sock, struct capture_device *cdev) {
	int c, ctrl_index = 0, value = 0, ret = ACTION_LIST;
	char *buf, *sptr, *fptr;
	struct control_list *l;
	
	XMALLOC(buf, char *, INPUT_BLOCK_SIZE);
	c = read(sock, buf, INPUT_BLOCK_SIZE - 1);
	buf[c] = '\0';
	
	if(strstr(buf, "webcam") != NULL) 
		ret = ACTION_CAPTURE;
	else if (strstr(buf, "list") != NULL){
		if((sptr = strstr(buf, "?")) != NULL) {
			fptr = strstr(sptr, " HTTP");
			*fptr = '\0';
			if(sscanf(++sptr, "val=%6d&%3d=update", &value, &ctrl_index) == 2) {
				//catch the jpeg control setting
				if(ctrl_index==-1) {
					info(LOG_INFO, "Setting JPEG quality to %d\n", value);
					if((1 <= value) && (value <= 100)) jpeg_quality=value;
					else info(LOG_ERR, "Invalid jpeg quality value %d\n", value);
				} else if(ctrl_index==-2) {
				//catch the frame ratio control
					info(LOG_INFO, "Setting frame ratio to %d\n", value);
					if((1 <= value) && (value <= 25)) { 
						requested_fps = value; set_fps(requested_fps);
					} else info(LOG_ERR, "Invalid frame rate %d\n", value);
				} else {
					l = cdev->ctrls;
					assert(ctrl_index < l->count);
					info(LOG_INFO, "Setting %s to %d\n", l->ctrl[ctrl_index].name, value);
					set_control_value(cdev, &l->ctrl[ctrl_index], value);
				}
			} else
				info(LOG_ERR, "Error parsing URl. Unable to set new value\n");
		}
	}
	XFREE(buf);
	return ret;
}

/* Starts a thread to handle a webcam connection on "sock" socket
 */
void start_thread_client(int sock, struct capture_device *cdev) {
	pthread_attr_t attr;
	pthread_t tid;
	struct thread_data *td;
	
	XMALLOC(td, struct thread_data *, sizeof(struct thread_data)); 
	if (td != NULL) {
		td->sock = sock;
		td->cdev = cdev;
		pthread_attr_init(&attr);
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
		if (pthread_create(&tid, &attr, send_stream_to, (void *)td) == 0) {
			client_nr++;
			dprint(LOG_SOURCE_HTTP, LOG_LEVEL_DEBUG2, "Created new thread \n");
		}
		else
			info(LOG_ERR, "Cant create a thread to handle the connection...\n");

		pthread_attr_destroy(&attr);
	}
	else
			info(LOG_ERR, "No more memory :(\n");
}

/* Webcam stream thread: v is a ptr to a struct thread_data which packages a socket and a cdev
 * this function is meant to run as a separate thread and handles a webcam stream connection 
 * on "sock" socket: It sends and HTTP header and then continuously captures a frame from cdev,
 * jpeg-encodes it and sends it. The loop is terminated when keep_going is false or when the 
 * socket is closed from the other end 
 */
void *send_stream_to(void *v) {
	struct v4l2_buffer *b;
	struct timeval start, now;
	struct timespec sleep_length, sleep_rem;
	struct jpeg j;
	struct thread_data *td = (struct thread_data *)v;
	struct capture_device *cdev = td->cdev;
	int yuv_len, jpeg_len, f_nr = 0, retval = 0, sock = td->sock; //, calibrated = 1;
	void *yuv_data, *jpeg_data;
	float ts, last_fps;
	
	XMALLOC(jpeg_data, void *, (cdev->width * cdev->height * 3));
	dprint(LOG_SOURCE_HTTP, LOG_LEVEL_DEBUG2, "New thread starting sock: %d, jpeg_quality: %d\n", sock, jpeg_quality);
	
	//setup jpeg param
	j.cinfo.err = jpeg_std_error(&j.jerr);
	jpeg_create_compress(&j.cinfo);
	j.cinfo.image_width = cdev->width;
	j.cinfo.image_height = cdev->height;
	j.cinfo.input_components = 3;
	jpeg_set_defaults(&j.cinfo) ;
	jpeg_set_colorspace(&j.cinfo, JCS_YCbCr);
	
	j.cinfo.raw_data_in = TRUE; // supply downsampled data
	j.cinfo.comp_info[0].h_samp_factor = 2;
	j.cinfo.comp_info[0].v_samp_factor = 2;
	j.cinfo.comp_info[1].h_samp_factor = 1;
	j.cinfo.comp_info[1].v_samp_factor = 1;
	j.cinfo.comp_info[2].h_samp_factor = 1;
	j.cinfo.comp_info[2].v_samp_factor = 1;
	
	jpeg_set_quality(&j.cinfo, jpeg_quality,TRUE);
	j.cinfo.dct_method = JDCT_FASTEST;
	
	j.destmgr.init_destination = init_destination;
	j.destmgr.empty_output_buffer = empty_output_buffer;
	j.destmgr.term_destination = term_destination;
		
	j.cinfo.dest = &j.destmgr;
	
	//send mjpeg header
	if (send_mjpeg_header(sock) != 0 )
		goto end;
	
	gettimeofday(&start, NULL);
	sleep_rem.tv_sec = sleep_rem.tv_nsec = 0;
		
	while((retval>=0) && keep_going) {
		
		gettimeofday(&now, NULL);
		if((now.tv_sec>=start.tv_sec + SHOW_FPS_INTERVAL)) {
			ts = (now.tv_sec - start.tv_sec) + ((float) (now.tv_usec - start.tv_usec)/1000000);
			last_fps = (f_nr / ts);
			//fprintf(stderr, "%d frames in %.3f sec (fps=%.1f, req_fps = %d)\n",f_nr, ts, last_fps, requested_fps);
			info(LOG_INFO, "%d frames in %.3f sec (fps=%.1f)\n",f_nr, ts, last_fps);
			
			//check whether the calibration was done
			if(fps_nanosleep_step == 0) {
				//no it wasnt, compute fps_nanosleep_step, the time it takes to capture,
				//jpeg-encode and send a single frame
				fps_nanosleep_step = (1000000000 / last_fps);
				dprint(LOG_SOURCE_HTTP, LOG_LEVEL_DEBUG1,"Calibrating loop: set nanosleep step to %09ld\n",fps_nanosleep_step);
				max_fps = last_fps;
				set_fps(requested_fps);
			} else {
				//calibration already completed
				//is current_fps = req_fps +- 0.5 ?
				if(last_fps<(requested_fps - 0.5 )) 
					decr_nanosleep();
				else if (last_fps >(requested_fps + 0.5 ))
				 	incr_nanosleep();
				else
					dprint(LOG_SOURCE_HTTP, LOG_LEVEL_DEBUG1,"Current fps_nanosleep (%ld.%09ld) achieves the desired framerate\n", fps_secsleep, fps_nanosleep);
			}
			
			f_nr = 0;
			gettimeofday(&start, NULL);
		}
		
		//sleep to adjust the fps to correct value
		if(((fps_nanosleep != 0) || (fps_secsleep != 0))) { 
			sleep_length.tv_nsec = fps_nanosleep;
			sleep_length.tv_sec = fps_secsleep;
			nanosleep(&sleep_length, &sleep_rem);
		}
	
		//get frame from v4l2 
		if((b = dequeue_buffer(cdev)) != NULL) {
			
			//get address of frame
			yuv_data = get_frame_buffer(cdev, b, &yuv_len);

			//encode in JPEG
			jpeg_encode(yuv_data, cdev->width, cdev->height, &jpeg_len, &j, jpeg_data);

			//return buffer to v4l2
			enqueue_buffer(cdev, b);
			
			//send in multipart_jpeg stream	
			retval = send_frame(sock, jpeg_data, jpeg_len);
				
			f_nr++;
		}
	}
	
	end:
	XFREE(jpeg_data);
	client_nr--;
	
	//close socket
	info(LOG_INFO, "closing connection on socket %d\n", sock );
	close(sock);
	
	XFREE(v);
	jpeg_destroy_compress(&j.cinfo);
	return NULL;
}
/* Sends a motion JPEG frame at "d" of size "len" over socket "sock"
 * returns 0 (success) or (-1) failure because of a TCP connection problem (ie, close)
 */
int send_frame(int sock, void *d, int len) {
	int  bs = 0, retval = 0;
	char c[14];
	
	//send the header	
	if(write(sock, MP_JPEG_FRAME_HEADER, strlen(MP_JPEG_FRAME_HEADER)) < 0) {
		dprint(LOG_SOURCE_HTTP, LOG_LEVEL_DEBUG1, "Error sending the MJPEG frame header\n");
		retval = -1;
		goto end;
	}	
	sprintf(c, "%d\r\n\r\n",len);
	if(write(sock, c, strlen(c))<0) {
		dprint(LOG_SOURCE_HTTP, LOG_LEVEL_DEBUG1, "Error sending the MJPEG frame header\n");
		retval = -1;
		goto end;
	}
	
	//sends the frame itself
	while(len > 0) {
		bs = (len < OUTPUT_BLOCK_SIZE) ? len : OUTPUT_BLOCK_SIZE;
		if(write(sock, d, bs) <0) {
			dprint(LOG_SOURCE_HTTP, LOG_LEVEL_DEBUG1, "Error sending the MJPEG frame\n");
			retval = -1;
			break;
		}
		len -= bs;
		d += bs; 
	}
	
	end:
	return retval;
}

/* Encodes a YUV planar frame of width "width and height "height" at "src" straight 
 * into a JPEG frame at "dst" (must be allocated y caller). "len" is set to the
 * length of the compressed JPEG frame. "j" contains the JPEG compressor and 
 * must be initialised correctly by the caller
 */
void jpeg_encode(void *src, int width, int height, int *len, struct jpeg *j, void *dst) {
	//Code for this function is taken from Motion
	//Credit to them !!!
	
	JSAMPROW y[16],cb[16],cr[16]; 
	JSAMPARRAY data[3]; 
	int i, line, rgb_size;
	
	dprint(LOG_SOURCE_JPEG, LOG_LEVEL_DEBUG2, "Encoding a %dx%d frame\n", width, height);
	
	data[0] = y;
	data[1] = cb;
	data[2] = cr;
	
	//init JPEG dest mgr
	rgb_size = width * height * 3;
	j->destmgr.next_output_byte = dst;
	j->destmgr.free_in_buffer = rgb_size;
	jpeg_set_quality(&j->cinfo, jpeg_quality,TRUE);
		
	jpeg_start_compress( &j->cinfo, TRUE );
	for (line=0; line<height; line+=16) {
		for (i=0; i<16; i++) {
			y[i] = src + width*(i+line);
			if (i%2 == 0) {
				cb[i/2] = src + width*height + width/2*((i+line)/2);
				cr[i/2] = src + width*height + width*height/4 + width/2*((i+line)/2);
			}
		}
		jpeg_write_raw_data(&j->cinfo, data, 16);
	}
	jpeg_finish_compress(&j->cinfo);
	*len = rgb_size - j->cinfo.dest->free_in_buffer;
	dprint(LOG_SOURCE_JPEG, LOG_LEVEL_DEBUG2, "Compressed %d bytes to %d bytes\n", rgb_size, *len);
}

/* Sends an MJPEG HTTP header over socket "sock"
 * returns 0(success) or -1(failure)
 */
int send_mjpeg_header(int sock) {
	char *header = NULL;
	int retval = 0;

	XMALLOC(header, char *, MP_JPEG_HTTP_HEADER_MAX_LENGTH + 1);
	sprintf(header, MP_JPEG_HTTP_HEADER, MP_JPEG_HTTP_BOUNDARY);

	if(write(sock, header, strlen(header)) < 0) {
		dprint(LOG_SOURCE_HTTP, LOG_LEVEL_DEBUG1, "Error sending the MJPEG header\n");
		retval = -1;
	}
	XFREE(header);
	return retval;
}

/*
 * Set fps_nanosleep and fps_secsleep according to a requested fps (int f)
 */
void set_fps (int f) {
	if(max_fps - requested_fps <= 0) {
		fps_nanosleep = 0;
		fps_secsleep = 0;
	} else {
		fps_secsleep = 0;
		fps_nanosleep = (1000000000 / requested_fps) - fps_nanosleep_step;
	}
	dprint(LOG_SOURCE_HTTP, LOG_LEVEL_DEBUG1,"Setting fps_nanosleep to %ld.%09ld\n", fps_secsleep, fps_nanosleep);
}

/* Increase fps{nano,sec}sleep by a fraction of fps_nanosleep_step
 * to slightly decrease the current fps
 */
void incr_nanosleep() {
	if((fps_nanosleep += (fps_nanosleep_step/8)) > 999999999) {
		fps_secsleep +=1;
		fps_nanosleep -= 1000000000;
	}
	dprint(LOG_SOURCE_HTTP, LOG_LEVEL_DEBUG1,"Incrementing fps_nanosleep %ld.%09ld\n", fps_secsleep, fps_nanosleep);
}

/* Decrease fps{nano,sec}sleep by a fraction of fps_nanosleep_step
 * to slightly decrease the current fps
 */
void decr_nanosleep() {
	if((fps_nanosleep -= (fps_nanosleep_step/8)) < 0) {
		fps_secsleep -=1;
		fps_nanosleep -= 1000000000;
	}
	if(fps_secsleep <0) {
		fps_secsleep = 0;
		fps_nanosleep = 0;
	}
	dprint(LOG_SOURCE_HTTP, LOG_LEVEL_DEBUG1,"Decrementing fps_nanosleep %ld.%09ld\n", fps_secsleep, fps_nanosleep);
}

void init_destination( j_compress_ptr cinfo )
{
}

boolean empty_output_buffer( j_compress_ptr cinfo )
{
	return TRUE;
}

void term_destination( j_compress_ptr cinfo )
{
}

void print_usage() {
	fprintf(stdout, "Usage: light_cap <device> <height> <width> <channel> <standard> <jpeg_quality> <fps> <verbosity>");
#ifdef DEBUG
	fprintf(stdout, " <log_source> <log_level>\n");
#else
	fprintf(stdout, "\n");
#endif
	fprintf(stdout, "All options are mandatory.\n");
	fprintf(stdout, "<device>: full path to the video device file\n");
	fprintf(stdout, "<height>: requested height of captured frames.\n");
	fprintf(stdout, "<width>: requested width of captured frames.\n");
	fprintf(stdout, "(The V4L2 layer will use the closest height and width to the ones requested if they are not available)\n");
	fprintf(stdout, "<channel>: capture channel to be used (0 for webcams).\n");
	fprintf(stdout, "<standard>: standard of the video device (0: webcam, 1: PAL, 2: SECAM, 3:NTSC).\n");
	fprintf(stdout, "<jpeg_quality>: quality (0 - 100) of generated JPEG images\n");
	fprintf(stdout, "<fps>: required frame rate (1 - 25)\n");
	fprintf(stdout, "<verbosity>: what to print on terminal: 0: nothing; 1: info; 2: errors; 3: both\n");
#ifdef DEBUG
	fprintf(stdout, "<log_source>: source of debug info (see log.h)\n");
	fprintf(stdout, "<log_level>: level of debug info (see log.h)\n");
#endif
	fprintf(stdout, "\nlight_cap is a lightweight video capture program. It grabbs frames from a V4L2 device and sends them\n");
	fprintf(stdout, "as a motion JPEG stream, which is accessible at the URL \"http://host:8080/webcam\"\n");
	fprintf(stdout, "light_cap also allows you to change capture parameters on the run via a set of web pages,\n");
	fprintf(stdout, "accessible at the URL \"http://host:8080/list\"\n");
}
